import signal
import time
from copy import deepcopy
from random import randint
from constants import English, Spanish, TerrainTypes
from games.cli_game import CliGame
from camels_race_exceptions import ErrorInConfig, InputTimedOut
from camel import CamelTypes
from utils import get_current_time_in_ms, json_load, pickle_list_to_file
from collections import namedtuple, OrderedDict
from operator import itemgetter

languages = {"english": English, "spanish": Spanish}
input_result = namedtuple("input_result", "turbo weapon steps")


class CamelRace(CliGame):
    # loads strings from the game's module to support basic internationalization
    name = "Camel Race"
    const_strings = languages["english"]

    # Drawing related class variables
    splash_screen = """
        @@@@#;',    @@@@
      @@@@@@@@@     @@@@@
      @@@@@@@.     @@@@@@
         @@@@`   .@@@@@@@@
         @@@@   @@@@@@@@@@@@
         @@@@  @@@@@@@@@@@@@.
         +@@@@@@@@@@@@@@@@@@@
          @@@@@@@@@@@@@@@@@@@
          `@@@@@@@@@@@@@@@@@#
            @@@@@@@@@@@@@@@@
              @@@@@@@@@@+@@@
             @@@@@@@@,   ,@@
            @@@  @@       ;@#
            @@   '@'       @@
            @+    @@       ,@:
            @@    @@       +@
            '@    @@       @@
             @@   @@       @@
             +    @'       @
                  @`      @@
    """
    screen_min_rows = 24
    screen_min_columns = 80
    finish_line = ("|F|", "|I|", "|N|")
    start_lines = (("C|", "P|", "U|"), ("Y|", "O|", "U|"))
    finish_line_width = len(finish_line[0])
    start_lines_width = len(start_lines[0][0])
    camel_width = 9
    side_lane = None
    winner_camel_position = screen_min_columns - finish_line_width - start_lines_width - camel_width
    lane_separator = "-" * screen_min_columns + "\n"
    visible_lane = screen_min_columns - finish_line_width - start_lines_width - camel_width

    banner = "\n" * 100
    banner += splash_screen
    banner += "\n          " + const_strings.WELCOME + "\n"

    def __init__(self, config, dump=False):
        """ create camels from config using factory
        if dump is True, saves a copy of state after each round
        """
        super(CamelRace, self).__init__(config)
        self.dump = dump
        self.states_to_dump = []  # useful for blackbox testing and debugging
        self.results_to_dump = [] # useful for blackbox testing and debugging
        self.T = self.config.T
        self.Tp = self.config.Tp
        self.X = self.config.X
        self.language = self.config.language
        self.terrain = self.config.terrain

        self.camels = self.config.camels  # ordered dict: camel name to camel object
        self.lane_2_camel = self.config.lane_2_camel  # direct mapping of lane to camel name
        self.results = {}  # holds input result per camel_name

        self.const_strings = languages[self.language]
        self.num_players = len(self.camels)
        self.user_io = InputProcessor(self.Tp, self.X)

        signal.signal(signal.SIGALRM, self.alarmHandler)

    def reset(self):
        for camel in self.camels.values():
            camel.reset()
        del self.messages[:]
        self.winner = None

    def get_state_drawing(self):
        """ draws race field, terrain and camels, only uses camels{}
        :return: screen: string that reflects race state
        """
        if not self.side_lane:  # change drawing according to terrain
            self.side_lane = TerrainTypes.icon[self.terrain] * self.screen_min_columns + "\n"
        screen = self.side_lane
        camels = self.camels.values()
        for camel_name, camel in self.camels.items():
            start_line_type = (1,0)[camel.player == "cpu"]
            for row in (0,1,2):  # draw row by row
                screen += (self.start_lines[start_line_type][row] +
                           " " * camel.position +  # leading space
                           camel.icon[row] +  # part of camel in this row
                           " " * (self.visible_lane - camel.position)   # space till finish line
                           + self.finish_line[row] + "\n")  # finish line
            # print line separator after each lane, except last one
            if camels.index(camel) != self.num_players-1:
                screen += self.lane_separator
        screen += self.side_lane
        return screen

    def alarmHandler(self, *args):
        raise InputTimedOut

    def next_turn(self):
        """ main logic driver of the race
        for each camel, gets input from player or cpu
            gets isolated result and get dequeue messages to display
        finally updates overall state by composing results from all camels
        """
        for camel in self.camels.values():
            if camel.player.lower() != "cpu":
                requested_input = self.user_io.get_challenge()  # get 3 letters string
                signal.alarm(self.T / 1000)
                start_time = get_current_time_in_ms()
                try:
                    user_input = raw_input(
                        "{}, {} [ {} ]:".format(
                        camel.player, CamelRace.const_strings.INPUT_TURN, requested_input))

                except InputTimedOut:
                    user_input = ''
                else:
                    signal.alarm(0)
                finish_time = get_current_time_in_ms()

                time_taken = finish_time - start_time
                result = self.user_io.process_input(requested_input, user_input, time_taken)

                #  comment following line to return to user after entering
                #  (no need to wait until T), it makes better game play experience IMHO
                time.sleep(abs(self.T - time_taken) / 1000) # only real players are paused
            else:
                result = self.user_io.get_random_result()
                if self.dump:
                    self.results_to_dump.append(deepcopy(result))

            self.queue_message(Commenter.get_comment(camel.name, result, self.const_strings))
            self.results[camel.name] = result

        # update positions after each round (for all camels)
        self.update_camels_positions()
        self.check_for_winner()
        if self.dump:
            self.dump_game_state()

    def dump_game_state(self):
        """ saves a copy of self.camels at the end of each round
        when we have a winner, saves the list to a json file"""
        self.states_to_dump.append(deepcopy(self.camels))
        if self.winner:  # save once game finished
            pickle_list_to_file("state.pkl",self.states_to_dump)
            pickle_list_to_file("results.pkl", self.results_to_dump)

    def update_camels_positions(self):
        """ updates camels positions taking into account camel's results
        and interaction with other camels"""

        # update chasers and forward lists of each camel
        self.calculate_relative_positions()

        # update received attacks of each camel
        for camel in self.camels.values():
            if self.results[camel.name].weapon:
                camel.use_weapon(self.camels)

        # Accumulate everything:
        #         new position = old position +
        #                        steps as result from input +
        #                        turbo_boost_effect +
        #                        attacks effect +
        #                        Strenght_reduction

        # I consider speed reduction = steps reduction
        for camel in self.camels.values():
            turbo_effect = 0  # positive percent
            attacks_effect = camel.absorb_attacks()  # negative percent and/or absolute value
            steps_from_input = self.results[camel.name].steps  # positive absolute value

            if self.results[camel.name].turbo:
                turbo_effect = camel.get_turbo_boost(self.terrain)
            final_percent = turbo_effect + camel.strength - attacks_effect["percent"]
            final_steps = steps_from_input - attacks_effect["absolute"]
            camel.position += int(final_steps * ((100 + final_percent) / 100.0))

    def calculate_relative_positions(self):
        """calculates relative positions and update camel.forward_camels
         and camel.chaser_camels for each camel in self.camels{}

        camels at the same position dont count as chaser nor forward
        'not the most pythonic method but it is late at night!'
        """

        # construct 2d array of position differences
        # each line represents differences between positions of other camels
        positions_differences = []
        for lane in xrange(self.num_players):
            positions_differences.append([])
            for other_camel in self.camels.values():
                positions_differences[lane].append(
                    self.lane_2_camel[lane].position - other_camel.position)

        lane = 0
        for diffs in positions_differences:
            # we need nearest to zero differences
            # get list of (camel index, relative diff to camel at that index)
            negative_diffs = [(i,diff) for i, diff in enumerate(diffs) if diff < 0]
            positive_diffs = [(i,diff) for i, diff in enumerate(diffs) if diff > 0]

            del self.lane_2_camel[lane].forward_camels[:]
            if negative_diffs:
                # get indexes with max negative value to find forward camels
                max_neg_diff = max(negative_diffs, key=itemgetter(1))[1]
                self.lane_2_camel[lane].forward_camels.extend(
                    [self.lane_2_camel[i].name for i,diff in negative_diffs if diff == max_neg_diff])

            del self.lane_2_camel[lane].chaser_camels[:]
            if positive_diffs:
                # get indexes with min positive value to find chaser camels
                min_pos_diff = min(positive_diffs, key=itemgetter(1))[1]
                self.lane_2_camel[lane].chaser_camels.extend(
                    [self.lane_2_camel[i].name for i,diff in positive_diffs if diff == min_pos_diff])
            lane +=1

    def check_for_winner(self):
        """ check if any camel reached finish line
        also checkes if tie has occurred
        modifies self.winner"""
        winners = []
        for camel in self.camels.values():
            if camel.position == self.winner_camel_position:
                winners.append(camel.name)

        if len(winners) > 1:
            self.winner = self.const_strings.TIE
        elif len(winners) == 1:
            self.winner = self.const_strings.WINNER.format(winners[0])


class InputProcessor(object):
    """ processes user input according to game rules, also simulates cpu players"""

    # use ascii codes to avoid different locales problems
    inputs = [chr(ascii_letter) for ascii_letter in xrange(65, 91)]
    inputs.extend(["I","U"]) # some extra turbo

    def __init__(self, Tp, X):
        self.Tp = Tp
        self.X = X

    @staticmethod
    def get_challenge():
        requested_input = ""
        for _ in (1, 2, 3):
            requested_input += InputProcessor.inputs[
                randint(0, len(InputProcessor.inputs)-1)]
        return requested_input

    def process_input(self, requested_input, user_input, time_taken):
        """ get the result of the input without taking into account other camels
        :param requested_input: string of three letters
        :param user_input: string of whatever user input
        :param time_taken: time consumed by player to enter
        :return: result:  namedtuple("input_result", "turbo weapon steps")
        """
        """ """
        user_input = user_input.upper()

        turbo = self.check_turbo(requested_input)
        weapon = False
        steps = 0

        if len(user_input) in (3, 6) and user_input[0:3] == requested_input:
            if time_taken <= self.Tp / 2:
                steps = self.X
            elif time_taken <= self.Tp:
                steps = self.X / 2

            # check if player stroked both input and reversed input within time T'
            if len(user_input) == 6 and steps != 0:
                if (user_input[0] == user_input[5] and
                            user_input[1] == user_input[4] and
                            user_input[2] == user_input[3]):
                    weapon = True

        return input_result(turbo=turbo, weapon=weapon, steps=steps)

    def get_random_result(self):
        """ get results when player is cpu,
        returns normalized random values to simulate humans
        :return: namedtuple("input_result", "turbo weapon steps")"""
        turbo = self.check_turbo(self.get_challenge())

        if self.Tp > 10000:
            # be fair with cpu, with 10s X is achievable even for 4004 :D
            steps = self.X
        else:
            steps = (self.X, self.X / 2, self.X / 2, self.X / 2, self.X / 2, 0)[randint(0, 5)]

        weapon = False
        if steps != 0:
            weapon = (True, False, False)[randint(0, 2)]

        return input_result(turbo=turbo, weapon=weapon, steps=steps)

    @staticmethod
    def check_turbo(requested_input):
        """
        turbo activation only depends on requested_input
        :param requested_input:
        :return: boolean reflecting if turbo was activated
        """
        if "I" in requested_input or "U" in requested_input:
            return True
        return False


class Commenter(object):
    """ used as code organizer for the comments related static method(s)"""

    @staticmethod
    def get_comment(name, result, strings):
        """
        :param name: camel name
        :param result:  namedtuple("input_result", "turbo weapon steps")
        :param strings: comment strings of one languages
        :return:
        """
        nice_word = strings.COMMENT_NICE_WORDS[randint(0, 3)]
        if result.weapon and result.turbo:
            comment = strings.COMMENT_WOW.format(name)
        elif result.turbo:
            comment = strings.COMMENT_TURBO.format(nice_word, name)
        elif result.weapon:
            comment = strings.COMMENT_WEAPON.format(nice_word, name)
        elif not randint(0, 2):  # if none applies, then third fo the time make another comment
            if randint(0, 1):  # can be generic or personalized per camel
                comment = strings.COMMENTS_RAND_PERS[randint(0, 2)].format(name)
            else:
                comment = strings.COMMENTS_RAND_GEN[randint(0, 8)]
        else:
            comment = None

        return comment


class CamelRaceGameConfig(object):
    """ parses and validates configs, instantiates camels as part of configs"""

    # DefaultConfigs applies when a key is missing, only in file mode as keys are
    # enforced in interactive mode
    DefaultConfigs = namedtuple("DefaultConfigs", "T TP X TERRAIN LANGUAGE")
    default_Configs = DefaultConfigs(T=6,
                                     TP=4,
                                     X=10,
                                     TERRAIN="Sand",
                                     LANGUAGE="English")

    def __init__(self, config):
        """
        Parse and validate configs
        instantiate camel objects and equip them with weapons
        :param config: json representation of global config and camels
        """

        try:
            global_config = config.get("global")
            self.T = 1000 * int(global_config.get("T", CamelRaceGameConfig.default_Configs.T))
            self.Tp = 1000 * int(global_config.get("Tp", CamelRaceGameConfig.default_Configs.TP))
            if self.T < self.Tp:
                raise ErrorInConfig("T must be bigger than Tp")
            self.X = int(global_config.get("X", CamelRaceGameConfig.default_Configs.X))
            self.terrain = global_config.get("terrain", CamelRaceGameConfig.default_Configs.TERRAIN)
            if self.terrain not in TerrainTypes.types:
                raise ErrorInConfig("Terrain {} not supported".format(self.terrain))
            self.language = global_config.get("language", CamelRaceGameConfig.default_Configs.LANGUAGE).lower()
            if self.language not in languages:
                raise ErrorInConfig("Not supported language {}".format(self.language))

            self.camels = OrderedDict()
            self.lane_2_camel = {}
            lane = 0
            for camel_config in config["camels"]:
                camel_name = camel_config.get("name")
                if camel_name in self.camels:
                    raise ErrorInConfig("Camel name must be unique, {} repeated".
                                         format(camel_name))

                self.camels[camel_name] = self.get_camel_from_json(camel_config)
                self.camels[camel_name].equip_weapon(impact=self.X)
                self.lane_2_camel[lane] = self.camels[camel_name]
                lane += 1

            self.num_players= len(self.camels)

        except (KeyError,ValueError) as ex:
            raise ErrorInConfig(ex.message)

    @staticmethod
    def get_camel_from_json(json_dict):
        """Camels factory method from json
        :raises keyError: raises an exception
        """
        camel_type = json_dict.get("type")
        if camel_type in CamelTypes.class_map:
            # return the corresponding camel object
            return CamelTypes.class_map[camel_type](
                json_dict.get("name"), json_dict.get("player").lower(),
                CamelRace.winner_camel_position)

    @staticmethod
    def prompt_for_config():
        """
        :return: CamelRaceGameConfig
        """

        config_option = raw_input("Enter 'C' for config mode or [enter] to load config from file:")
        if config_option.lower() != "c":
            configs = json_load(raw_input("Enter config file name [config.json]:").strip() or "config.json")
        else:
            try:
                num_camels = int(raw_input("Enter number of camels (minimum 1) default [3]:") or "3")
                if num_camels <= 0:
                    num_camels = 1
                camels = []
                for i in xrange(num_camels):
                    camel = {}
                    camel["name"] = raw_input(
                        "[config camel {}] Enter camel name:".format(i)).strip()  or "camel_{}".format(i+1)
                    camel["type"] = raw_input(
                        "[config camel {}] Enter camel type '[Bactrian] Domestic Dromedary':".format(i)).strip()  or "Bactrian"
                    camel["player"] = raw_input(
                        "[config camel {}] Enter player name or [cpu]:".format(i)).strip()  or "cpu"
                    camels.append(camel)
                global_configs = {}
                global_configs["T"] = abs(int(raw_input("Enter T [6]:") or "6"))
                global_configs["Tp"] = abs(int(raw_input("Enter Tp [4]:") or "4"))
                global_configs["X"] = abs(int(raw_input("Enter X [10]:") or "10"))
                global_configs["terrain"] = raw_input("Enter terrain type '[Sand] Mud Grass':").strip() or "Sand"
                global_configs["language"] = raw_input("Enter UI language '[English] Spanish':").strip() or "English"

                configs = {"global":global_configs, "camels":camels}
            except ValueError as ex:
                raise ErrorInConfig("Error Getting config from user:{}".format(ex.message))

        return CamelRaceGameConfig(configs)

    def __eq__(self, other):
        globals_are_equal = (self.T == other.T and
        self.Tp == other.Tp and
        self.X == other.X and
        self.language == self.language and
        self.terrain == other.terrain)

        camles_are_equal = True
        for camel_name in self.camels.keys():
            if camel_name in other.camels:
                if not (self.camels[camel_name] == other.camels[camel_name]):
                    camles_are_equal = False
                    break
            else:
                camles_are_equal = False
                break

        return globals_are_equal and camles_are_equal


import time
import json
import cPickle as pickle
import os
import sys


def get_current_time_in_ms():
    return int(round(time.time() * 1000))


def json_load(filename):
    full_path = os.path.join(os.path.dirname(__file__), filename)
    try:
        with open(full_path) as data_file:
            data = json.load(data_file)
            return data
    except IOError as ex:
        print "Error loading json file '{}': {}".format(full_path, ex.message)
        sys.exit(1)


def pickle_list_to_file(filename, list_of_objects):
    full_path = os.path.join(os.path.dirname(__file__), filename)
    try:
        with open(full_path, "w+") as data_file:
            pickle.dump(list_of_objects, data_file, pickle.HIGHEST_PROTOCOL)
    except IOError as ex:
        print "Error saving to pickle file '{}': {}".format(full_path, ex.message)
        sys.exit(1)


def unpickle_list_from_file(filename):
    full_path = os.path.join(os.path.dirname(__file__), filename)
    try:
        with open(full_path, "r+") as data_file:
            return pickle.load(data_file)
    except IOError as ex:
        print "Error loading list from pickle file '{}': {}".format(full_path, ex.message)
        sys.exit(1)


from abc import ABCMeta, abstractmethod, abstractproperty


class CliGame(object):
    __metaclass__ = ABCMeta

    def __init__(self, config):
        self.config = config
        self.winner = None
        self.num_players = 0
        self._messages = []  # accumulates messages to print to console

    def queue_message(self, value):
        if value:  # ignore Nones
            self._messages.append(value)

    @property
    def messages(self):
        return self._messages

    @abstractproperty
    def screen_min_rows(self):pass

    @abstractproperty
    def screen_min_columns(self):pass

    @abstractproperty
    def banner(self):pass

    @abstractproperty
    def const_strings(self):pass

    @abstractproperty
    def name(self):pass

    @abstractmethod
    def next_turn(self):pass

    @abstractmethod
    def get_state_drawing(self):
        """ Returns a string that represents the internal state.
            Could be printed directly by the launcher"""
        pass


    @abstractmethod
    def reset(self):
        """ reset to initial state, conserves config"""
        pass


"""Contains Strings and possibly another constants"""

class English():
    WELCOME = "Welcome To Camel Race!"
    START_GAME = "Ready To Start An Epic Adventure? Press Any Key to Start:"
    INPUT_TURN = "Enter"
    EXIT = "Bye Bye!"
    COMMENT_WOW = "Wow, '{}' activates Turbo and uses weapon"
    COMMENT_TURBO = "{}, '{}' activates Turbo"
    COMMENT_WEAPON =  "{}, '{}' uses weapon"
    COMMENT_NICE_WORDS = ["Nice", "Cool", "Look!","Awsome!"]
    COMMENTS_RAND_PERS = ["'{}' looks tired today",
                          "'{}' is one of the best cammels",
                          "'{}' is working hard"]
    COMMENTS_RAND_GEN = [".....terrain is not helping ",
                         "...nice day for camel races",
                         "Typical spanish race!....",
                         "..What a great number of people watching",
                         "I love camels",
                         "Camels can survive 7 months without water!",
                         "...really cute animals",
                         "..greeting from the stadium..",
                         "These type of races move a lot of money!.."]
    TIE = "TIE!"
    WINNER = "The winner is {}, Congratulations!!!"
    PLAY_AGAIN = "Hit any button to play again, 'n' to exit:"


class Spanish():
    WELCOME = "Bienvenid@ a Camel Race!"
    START_GAME = "Listo Para Emprender Una Aventura Epica? Pulsa Para Empezar:"
    INPUT_TURN = "Entra"
    EXIT = "Adios!"
    COMMENT_WOW = "yuju, '{}' activa el turbo y usa su arma"
    COMMENT_TURBO = "{}, '{}' activa el turbo "
    COMMENT_WEAPON =  "{}, '{}' usa su arma"
    COMMENT_NICE_WORDS = ["Guay", "Como Mola!", "Fijate!","Genial!"]
    COMMENTS_RAND_PERS = ["'{}' parece cansado hoy",
                          "'{}' es uno de los mejores camellos",
                          "'{}' esta trabajando duro"]
    COMMENTS_RAND_GEN = [".....el suelo no ayuda",
                         "...buen dia para carrera de camellos",
                         "Tipico en los pueblos!....",
                         "..Cuanta gente viendo!",
                         "Adoro a los camellos",
                         "Un camellos puede vivir 7 meses sin agua!",
                         "...Que monos los camellos",
                         "..saludos desde el estadio..",
                         "Estas carreras mueven mucha pasta!.."]
    TIE = "EMPATE!"
    WINNER = "El Ganador es {}, Enhorabuena!!!"
    PLAY_AGAIN = "Pulsa cualquier tecla para jugar otra vez, 'n' para salir:"


class TerrainTypes(object):
    """ Enum like class for terrain types"""
    types = ["Mud", "Grass", "Sand"]
    Mud, Grass, Sand = types
    icon = {Mud:"~", Grass:"/", Sand:"="}


import sys
import signal
import subprocess
import time
from random import randint, sample
from games_exceptions import ErrorGettingConsoleDimensions
from games_exceptions import ConsoleDimenisionsTooSmall


class Launcher(object):
    """ Handles :
    1- the lifecycle of a CliGame
    2- handles all printing, prints game's splash screen, state and messages from game
    3- catches keyboard interrupts to exit game
    4- loops for player turns
    input is delegated to game via next_turn() method
    """
    max_popup_messages = 3

    def __init__(self, game):
        """
        Launches the game: initialize Launcher, print banner, prompt to start
        :param game: CliGame object
        """
        try:
            console_rows, console_columns = Launcher.get_console_dimensions()
            if console_columns < game.screen_min_columns or console_rows < game.screen_min_rows:
                raise ConsoleDimenisionsTooSmall("For '{}', min {}x{}".
                      format(game.name, game.screen_min_rows, game.screen_min_columns))
        except ErrorGettingConsoleDimensions:
            # not critical, we can live with it
            print ("Couldn't verify console dimensions, recommended min {}x{}".
                format(game.screen_min_rows,game.screen_min_columns))

        self.exit = False
        self.set_exit_handler()
        self.game = game
        if not self.game.dump:  # dont interrupt if dump mode is on
            self.print_banner()
            #  wait for user input to start the Game
            raw_input(self.game.const_strings.START_GAME)

    @staticmethod
    def clean_screen():
        print 100 * "\n"

    @staticmethod
    def get_console_dimensions():
        try:
            output = subprocess.check_output(["stty","size"])
            rows, columns = output.split()
            return int(rows), int(columns)
        except (OSError,ValueError) as ex:
            raise ErrorGettingConsoleDimensions(ex.message)

    def print_banner(self):
        for x in xrange(3):
            time.sleep(0.5)
            self.clean_screen()
            time.sleep(0.5)
            print self.game.banner

    def loop(self):
        self.clean_screen()
        while not self.exit and not self.game.winner:
            print "\n" + self.game.get_state_drawing()
            self.print_pending_messages()
            self.game.next_turn()
            self.clean_screen()
        print "\n" + self.game.get_state_drawing()
        self.print_popup([self.game.winner])
        if not self.game.dump:
            if self.play_again() :
                self.game.reset()
                self.loop()

    def play_again(self):
        user_input = raw_input(self.game.const_strings.PLAY_AGAIN)
        if user_input.lower() == "n":
            return False
        return True

    def print_pending_messages(self):
        """ Check if game wants to display messages"""
        if self.game.messages:
            if len(self.game.messages) > self.max_popup_messages:
                # limit popup messages
                self.print_popup(sample(self.game.messages, self.max_popup_messages))
            else:
                self.print_popup(self.game.messages)

            del self.game.messages[:]

    @staticmethod
    def print_popup(messages, left=None):
        """
        Formats messages into popup message then prints it
        :param messages: list of strings
        :param left: optional, determine which side is the popup
        >>> Launcher.print_popup(["msg1","msg1","msg2"], left=True)
        ,------.
        | msg1  |
        | msg2  |
        |/------
        <BLANKLINE>

        >>> Launcher.print_popup(["msg1","msg1","msg2"], left=False)
        ,------.
        | msg1  |
        | msg2  |
         ------\|
        <BLANKLINE>

        """

        max_popup_length = len(max(messages, key=len)) + 2
        top_boarder = "," + ("-" * max_popup_length) + "."
        if not left:  # if not specified, alter left and right popups
            left = randint(0, 1)
        if left:
            bottom_boarder = "|/" + ("-" * max_popup_length)
        else:
            bottom_boarder = " " + ("-" * max_popup_length) + "\|"

        popup = top_boarder + "\n"

        for msg in set(messages):
            popup += ("| " + msg + (" " * (max_popup_length - len(msg))) + "|") + "\n"
        popup += bottom_boarder + "\n"
        print popup

    def set_exit_handler(self):
        """ catch ctrl-c to exit cli game"""
        signal.signal(signal.SIGINT, self.catch_ctrl_c)

    def catch_ctrl_c(self, *args):
        print 2 * "\n" + self.game.const_strings.EXIT
        self.exit = True
        sys.exit(0)




from abc import ABCMeta, abstractproperty
from constants import TerrainTypes
from camel_weapon import CamelWeapon, AffectCamels, EffectType, WeaponTakesEffect


class Camel(object):
    __metaclass__ = ABCMeta

    def __init__(self, name, player, max_position):
        self._position = 0
        self.max_position = max_position
        self.name = name
        self.player = player
        self.received_attacks = []
        self.forward_camels = []  # camels names immediately forward
        self.chaser_camels = []  # camels names immediately chasing
        self.weapon = None  # will be created by equip_weapon()

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, value):
        """ ensure camel stays within race lanes limits"""
        if value > self.max_position:
            self._position = self.max_position
        elif value < 0:
            self._position = 0
        else:
            self._position = value

    @abstractproperty
    def icon(self):
        """ 3 rows * 9 columns """
        # we could check icons provided by son classes are 3x9
        # at __init__ of Camel object
        pass

    @abstractproperty
    def strength(self): pass

    @abstractproperty
    def default_boost(self):pass

    @abstractproperty
    def terrain_boost(self):pass

    @abstractproperty
    def favorite_terrain(self):pass

    def get_turbo_boost(self, terrain):
        if terrain == self.favorite_terrain:
            return self.terrain_boost
        else:
            return self.default_boost

    @abstractproperty
    def weapon_name(self):pass

    @abstractproperty
    def weapon_impact(self):pass

    @abstractproperty
    def weapon_affects(self):pass

    @abstractproperty
    def weapon_takes_affect(self):pass

    @abstractproperty
    def weapon_effect_type(self):pass

    def reset(self):
        self.position = 0
        self.chaser_camels = []
        self.forward_camels = []
        self.received_attacks = []

    def equip_weapon(self, impact=None):
        """ camels almost have all the info to create their weapons
        but we may need external data such as X from game object
        only overwrite value is self.wepon_impact is None"""
        if impact and not self.weapon_impact:
            self.weapon_impact = impact

        self.weapon = CamelWeapon(self.weapon_name,
                                  self.weapon_impact,
                                  affects=self.weapon_affects,
                                  takes_effect=self.weapon_takes_affect,
                                  effect_type=self.weapon_effect_type)

    def use_weapon(self, camels):
        if self.weapon.affects == AffectCamels.BOTH:
            for camel_name, camel in camels.items():
                if camel_name in self.chaser_camels + self.forward_camels:
                    camel.received_attacks.append(self.weapon)
        elif self.weapon.affects == AffectCamels.FORWARD:
            for camel_name, camel in camels.items():
                if camel_name in self.forward_camels:
                    camel.received_attacks.append(self.weapon)
        elif self.weapon.affects == AffectCamels.CHASER:
            for camel_name, camel in camels.items():
                if camel_name in self.chaser_camels:
                    camel.received_attacks.append(self.weapon)

    def absorb_attacks(self):
        """ returns accumulated effect of attacks from weapon
        with immediate effect, and delete theses attacks.
        also, change attacks with effect on next round
        to WeaponTakesEffect.now so they will be absorbed on the next round

        returns dict with accumulated effect per type
        """
        effect ={"absolute": 0 ,
                 "percent":0}

        for weapon in self.received_attacks:
            if weapon.takes_effect == WeaponTakesEffect.NOW:
                if weapon.effect_type == EffectType.FIXED:
                    effect["absolute"]+= weapon.impact
                else:
                    effect["percent"] += weapon.impact
                self.received_attacks.remove(weapon)

            else:
                weapon.takes_effect = WeaponTakesEffect.NOW

        return effect

    def __eq__(self, other):
        return (self.name == other.name and
                self.player == other.player and
                self.weapon == other.weapon)

    def __str__(self):
        weapon_name = "not equiped"
        if self.weapon:
            weapon_name = self.weapon.name
        return ("Camel: [name:{}, player:{}, position:{},\n "
                "   immediate forward{} chasers{}\n "
                "   recieved attacks{}"
                "   weapon {}]".format(
            self.name, self.player, self.position,
            self.forward_camels, self.chaser_camels,
            [a.name for a in self.received_attacks],
            weapon_name))


class BactrianCamel(Camel):
    strength = 20
    default_boost = 15
    terrain_boost = 30
    favorite_terrain = TerrainTypes.Mud
    weapon_name = "kick"
    weapon_impact = None # X should be set externally
    weapon_affects = AffectCamels.CHASER
    weapon_takes_affect = WeaponTakesEffect.NOW
    weapon_effect_type = EffectType.FIXED

    icon =(" /^^\ oo=",
           "+------' ",
           " |###|   ")

    def __init__(self,name, player, max_position):
        super(BactrianCamel, self).__init__(name, player,max_position)


class DomesticCamel(Camel):
    strength = 10
    default_boost = 20
    terrain_boost = 30
    favorite_terrain = TerrainTypes.Grass
    weapon_name = "sound"
    weapon_impact = 50
    weapon_affects = AffectCamels.BOTH
    weapon_takes_affect = WeaponTakesEffect.NEXT_ROUND
    weapon_effect_type = EffectType.PERCENT
    icon = (" /^^\ oo=",
            "-------' ",
            " ]---[   ")

    def __init__(self,name, player, max_position):
        super(DomesticCamel, self).__init__(name, player, max_position)


class DromedaryCamel(Camel):
    strength = 0
    default_boost = 25
    terrain_boost = 35
    favorite_terrain = TerrainTypes.Sand
    weapon_name = "spit"
    weapon_impact = 100  # That camel will not move forward during the next round
    weapon_affects = AffectCamels.FORWARD
    weapon_takes_affect = WeaponTakesEffect.NEXT_ROUND
    weapon_effect_type = EffectType.PERCENT
    icon = ( "  /^\ oo=",
             ".------' ",
             " |   |   ")

    def __init__(self,name, player, max_position):
        super(DromedaryCamel, self).__init__(name, player, max_position)


class CamelTypes(object):
    """ Enum like with class mapping for camel types"""
    Bactrian, Domestic, Dromedary = ["Bactrian", "Domestic", "Dromedary"]
    class_map = {Bactrian: BactrianCamel, Domestic: DomesticCamel, Dromedary: DromedaryCamel}



#!/usr/bin/env python
import sys
from games.camels_race.race_game import CamelRace, ErrorInConfig, CamelRaceGameConfig
from games.games_exceptions import ConsoleDimenisionsTooSmall
from games.cli_game_launcher import Launcher


def main():
    try:
        # ask user to enter config or load from file
        game_config = CamelRaceGameConfig.prompt_for_config()
        # create game
        game = CamelRace(game_config)
        # Launch game using console launcher
        cli_game_launcher = Launcher(game)
        cli_game_launcher.loop()

    except ConsoleDimenisionsTooSmall as ex:
        print "Console is too small :", ex.message
        sys.exit(-1)
    except ErrorInConfig as ex:
        print "Configuration Error :", ex.message

if __name__ == '__main__':
    main()



from enum import Enum

WeaponTakesEffect = Enum("WeaponTakesEffect", "NOW NEXT_ROUND")

AffectCamels = Enum("AffectCamels", "CHASER FORWARD BOTH")

EffectType = Enum("EffectType", "FIXED PERCENT")


class CamelWeapon(object):
    def __init__(self, name, impact,
                 affects=AffectCamels.BOTH,
                 takes_effect=WeaponTakesEffect.NOW,
                 effect_type=EffectType.PERCENT):
        self.name = name
        self.impact = impact
        self.affects = affects
        self.takes_effect = takes_effect
        self.effect_type = effect_type

    def __str__(self):
        return "Weapon:[name:{}, {}, {}, impact:{} {}]".format(
            self.name,
            self.affects,
            self.takes_effect,
            self.impact,
            self.effect_type)

    def __eq__(self, other):
        return self.name == other.name


#!/usr/bin/env python

from setuptools import setup, find_packages

setup(name='camel_race',
      version='1.0',
      description='Camel Race ascii game',
      author='Ibrahim Menem',
      author_email='ibrahimmenem@gmail.com',
      url='https://dummystack.com',
      packages=find_packages(),
      install_requires=['enum34'],
      data_files = [('./games/camels_race', ['config.json'])],
      entry_points={
        'console_scripts': [
            'camel_race=games.play:main',
        ],
       }

     )
